#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy
import functools
import io
import json
import logging
import tarfile

import requests

from atlasclient import models, utils, base, exceptions
from atlasclient.exceptions import handle_response

LOG = logging.getLogger(__name__)
LOG.addHandler(utils.NullHandler())

# this defines where the Atlas client delegates to for actual logic
ENTRY_POINTS = {'entity_guid': models.EntityGuid,
                'typedefs': models.TypeDef,
                'entity_post': models.EntityPost,
                'entity_bulk': models.EntityBulk,
                'entity_bulk_classification': models.EntityBulkClassification,
                'entity_unique_attribute': models.EntityUniqueAttribute,
                'typedefs_headers': models.TypeDefHeader,
                'classificationdef_guid': models.ClassificationDefGuid,
                'classificationdef_name': models.ClassificationDefName,
                'entitydef_guid': models.EntityDefGuid,
                'entitydef_name': models.EntityDefName,
                'enumdef_guid': models.EnumDefGuid,
                'enumdef_name': models.EnumDefName,
                'relationshipdef_guid': models.RelationshipDefGuid,
                'relationshipdef_name': models.RelationshipDefName,
                'structdef_guid': models.StructDefGuid,
                'structdef_name': models.StructDefName,
                'typedef_guid': models.TypeDefGuid,
                'typedef_name': models.TypeDefName,
                'lineage_guid': models.LineageGuid,
                'search_attribute': models.SearchAttribute,
                'search_basic': models.SearchBasic,
                'search_dsl': models.SearchDsl,
                'search_fulltext': models.SearchFulltext,
                'relationship': models.Relationship,
                'relationship_guid': models.RelationshipGuid,
                'search_saved': models.SearchSaved,
                'admin_metrics': models.AdminMetrics
               }


class Atlas(object):
    """The Atlas client

    This is the entry point to the Atlas API. Create this client and then
    use one of the entry points to start hitting Atlas object collections.
    """
    def __init__(self, host, port=None, username=None, password=None,
                 identifier=None, protocol=None, validate_ssl=True,
                 timeout=10, max_retries=5, auth=None):

        self.base_url = utils.generate_base_url(host, port=port, protocol=protocol)

        if identifier is None:
            identifier = 'python-atlasclient'

        self.client = HttpClient(host=self.base_url, username=username,
                                 password=password, identifier=identifier,
                                 validate_ssl=validate_ssl, timeout=timeout,
                                 max_retries=max_retries, auth=auth)
        self._version = None

    def __dir__(self):
        d1 = {}
        d1.update(self.__dict__)
        d1.update(ENTRY_POINTS)
        return d1.keys()

    def check_version(self):
        if self.version < base.OLDEST_SUPPORTED_VERSION:
            raise exceptions.ClientError(
                "Version %s unsupported, must be %s or higher"
                % (utils.version_str(self.version),
                   utils.version_str(base.OLDEST_SUPPORTED_VERSION)))
        return

    def __getattr__(self, attr):
        if attr in ENTRY_POINTS:
            rel_class = ENTRY_POINTS[attr]
            return rel_class.collection_class(self, rel_class)

        if getattr(requests, attr):
            # forward get/post/put/head/delete to the http client
            return getattr(self.client, attr)

        raise AttributeError(attr)


class HttpClient(object):
    """Our HTTP based REST client.

    It handles some of the dirty work like automatic serialization/deserialization
    of JSON data, converting error responses to exceptions, etc.  For the most
    part it should mimic a requests client. You can call methods like get, post,
    put, delete, and head and expect them to work the same way.  But instead of
    a response object, you get a dictionary.  A response of None means no response
    was supplied by the API.  This should be uncommon except for error cases, but
    cases do exist either due to Atlas bugs or other mitigating circumstances.
    """
    def __init__(self, host, username, password, identifier, validate_ssl=True,
                 timeout=10, max_retries=5, auth=None):
        basic_token = utils.generate_http_basic_token(username=username, password=password)
        self.request_params = {
            'headers': {'X-Requested-By': identifier,
                        'Authorization': 'Basic {}'.format(basic_token)},
            #'auth': (username, password),
            'verify': validate_ssl,
            'timeout': timeout,
        }
        # automatically retry requests on connection errors
        self.session = requests.Session()
        self.session.auth = auth
        adapter = requests.adapters.HTTPAdapter(max_retries=max_retries)
        self.session.mount(host, adapter)

    def request(self, method, url, content_type=None, **kwargs):
        # doing it this way keeps the magic for following redirects intact
        requests_method = getattr(self.session, method)
        params = copy.deepcopy(self.request_params)
        params.update(kwargs)

        if content_type is not None:
            params['headers']['Content-type'] = content_type
        else:
            params['headers']['Content-type'] = 'application/json'
        LOG.debug("Request headers: %s", params['headers'])

        if 'data' in params and isinstance(params['data'], dict):
            params['data'] = json.dumps(params['data'], cls=AtlasJsonEncoder)
            LOG.debug("Request body: %s", params['data'])
        elif 'data' in params and isinstance(params['data'], str):
            params['data'] = json.dumps(params['data'])
        elif 'data' in params and isinstance(params['data'], list):
            params['data'] = json.dumps(params['data'])

        response = requests_method(url, **params)

        # any error responses will generate exceptions here
        handle_response(response)

        LOG.debug("Response headers: %s", response.headers)
        LOG.debug("Response: %s", response.text)

        if response.headers.get('content-length') is None:
            # Log bad methods so we can report them
            LOG.debug("Missing content-length for %s %s: %s", method,
                      url, response.headers.get('content-type'))

        # there is no consistent way to determine response type
        # so assume json if it's not an empty string
        if response.text:
            if response.headers.get('content-type') == 'application/x-ustar':
                tarstream = io.BytesIO(response.content)
                tarstream.seek(0)
                return tarfile.open(fileobj=tarstream)
            elif 'application/json' not in response.headers.get('content-type'):
                # Log bad methods so we can report them
                LOG.debug("Wrong response content-type for %s %s: %s", method,
                          url, response.headers.get('content-type'))
            return response.json()

        return {}

    def __getattr__(self, attr):
        if getattr(requests, attr):
            return functools.partial(self.request, attr)
        raise AttributeError(attr)


class AtlasJsonEncoder(json.JSONEncoder):
    """Converts Atlas model objects into dictionaries that can be JSON-encoded

    This allows for passing in models and ModelCollections into related objects'
    create/update methods and having it handle the conversion automatically.
    """
    def default(self, obj):  # pylint: disable=method-hidden
        if isinstance(obj, base.ModelCollection):
            dicts = []
            for model in obj:
                dicts.append(model.to_json_dict())
            return dicts
        elif isinstance(obj, base.Model):
            return obj.to_json_dict()
        # Let the base class default method raise the TypeError
        return super(AtlasJsonEncoder, self).default(obj)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import re
import base64

try:
    from logging import NullHandler  # pylint: disable=unused-import
except ImportError:
    from logging import Handler


    class NullHandler(Handler):
        def emit(self, record):
            pass

DEFAULT_PORTS = {
    'http': 80,
    'https': 443,
}

DEFAULT_TABLE_QN_REGEX = re.compile(r"""
    ^(?P<db_name>.*?)\.(?P<table_name>.*)@(?P<cluster_name>.*?)$
    """, re.X)

DEFAULT_DB_CLUSTER = 'default'


def normalize_underscore_case(name):
    """Normalize an underscore-separated descriptor to something more readable.

    i.e. 'NAGIOS_SERVER' becomes 'Nagios Server', and 'host_components' becomes
    'Host Components'
    """
    normalized = name.lower()
    normalized = re.sub(r'_(\w)',
                        lambda match: ' ' + match.group(1).upper(),
                        normalized)
    return normalized[0].upper() + normalized[1:]


def normalize_camel_case(name):
    """Normalize a camelCase descriptor to something more readable.

    i.e. 'camelCase' or 'CamelCase' becomes 'Camel Case'
    """
    normalized = re.sub('([a-z])([A-Z])',
                        lambda match: ' '.join([match.group(1), match.group(2)]),
                        name)
    return normalized[0].upper() + normalized[1:]


def version_tuple(version):
    """Convert a version string or tuple to a tuple.

    Should be returned in the form: (major, minor, release).
    """
    if isinstance(version, str):
        return tuple(int(x) for x in version.split('.'))
    elif isinstance(version, tuple):
        return version
    else:
        raise ValueError("Invalid version: %s" % version)


def version_str(version):
    """Convert a version tuple or string to a string.

    Should be returned in the form: major.minor.release
    """
    if isinstance(version, str):
        return version
    elif isinstance(version, tuple):
        return '.'.join([str(int(x)) for x in version])
    else:
        raise ValueError("Invalid version: %s" % version)


def generate_http_basic_token(username, password):
    """
    Generates a HTTP basic token from username and password

    Returns a token string (not a byte)
    """
    token = base64.b64encode('{}:{}'.format(username, password).encode('utf-8')).decode('utf-8')
    return token


def generate_base_url(host, protocol=None, port=None):
    matches = re.match(r'^(([^:]+)://)?([^/:]+)(:([^/]+))?', host)
    (_, derived_proto, derived_host, _, derived_port) = matches.groups()
    if derived_proto is None:
        derived_proto = protocol or 'http'
    if derived_proto not in DEFAULT_PORTS:
        raise ValueError()

    if derived_port is None:
        derived_port = port or DEFAULT_PORTS[derived_proto]

    derived_port = int(derived_port)

    url_params = {
        'protocol': derived_proto,
        'host': derived_host,
        'port': str(derived_port),
    }
    return "{protocol}://{host}:{port}".format(**url_params)


def parse_table_qualified_name(qualified_name, qn_regex=DEFAULT_TABLE_QN_REGEX):
    """
    Parses the Atlas' table qualified name
    :param qualified_name: Qualified Name of the table
    :return: A dictionary consisting of database name,
    table name and cluster name of the table.
    If database or cluster name not found,
    then uses the 'atlas_default' as both of them.
    """
    def apply_qn_regex(name, table_qn_regex):
        return table_qn_regex.match(name)

    _regex_result = apply_qn_regex(qualified_name, qn_regex)

    if not _regex_result:
        qn_regex = re.compile(r"""
        ^(?P<table_name>.*)@(?P<cluster_name>.*?)$
        """, re.X)
        _regex_result = apply_qn_regex(qualified_name, qn_regex)

    if not _regex_result:
        qn_regex = re.compile(r"""
        ^(?P<db_name>.*?)\.(?P<table_name>.*)$
        """, re.X)
        _regex_result = apply_qn_regex(qualified_name, qn_regex)

    if not _regex_result:
        qn_regex = re.compile(r"""
        ^(?P<table_name>.*)$
        """, re.X)
        _regex_result = apply_qn_regex(qualified_name, qn_regex)

    _regex_result = _regex_result.groupdict()

    qn_dict = {
        'table_name': _regex_result.get('table_name', qualified_name),
        'db_name': _regex_result.get('db_name', DEFAULT_DB_CLUSTER),
        'cluster_name': _regex_result.get('cluster_name', DEFAULT_DB_CLUSTER),
    }

    return qn_dict


def make_table_qualified_name(table_name, cluster=None, db=None):
    """
    Based on the given parameters, generate the Atlas' table qualified Name
    :param db: Database Name of the table
    :param table_name: Table Name
    :param cluster: Cluster Name of the table
    :return: A string i.e., Qualified Name of the table
    If database or cluster name are 'atlas_default', then simply strips that part.
    """
    qualified_name = table_name
    if db and db != DEFAULT_DB_CLUSTER:
        qualified_name = '{}.{}'.format(db, qualified_name)

    if cluster and cluster != DEFAULT_DB_CLUSTER:
        qualified_name = '{}@{}'.format(qualified_name, cluster)

    return qualified_name



from atlasclient.utils import (parse_table_qualified_name, make_table_qualified_name,
                               DEFAULT_DB_CLUSTER)

DB = 'database_name'
CL = 'cluster_name'
TB = 'table_name'


class TestUtils():
    def test_parse_table_qn(self):
        qualified_name = '{}.{}@{}'.format(DB, TB, CL)
        qn_dict = parse_table_qualified_name(qualified_name)
        assert qn_dict['db_name'] == DB
        assert qn_dict['cluster_name'] == CL
        assert qn_dict['table_name'] == TB

    def test_parse_table_qn_without_db(self):
        qualified_name = '{}@{}'.format(TB, CL)
        qn_dict = parse_table_qualified_name(qualified_name)
        assert qn_dict['db_name'] == DEFAULT_DB_CLUSTER
        assert qn_dict['cluster_name'] == CL
        assert qn_dict['table_name'] == TB

    def test_parse_table_qn_without_cluster(self):
        qualified_name = '{}.{}'.format(DB, TB)
        qn_dict = parse_table_qualified_name(qualified_name)
        assert qn_dict['db_name'] == DB
        assert qn_dict['cluster_name'] == DEFAULT_DB_CLUSTER
        assert qn_dict['table_name'] == TB

    def test_parse_table_qn_only_table(self):
        qualified_name = '{}'.format(TB)
        qn_dict = parse_table_qualified_name(qualified_name)
        assert qn_dict['db_name'] == DEFAULT_DB_CLUSTER
        assert qn_dict['cluster_name'] == DEFAULT_DB_CLUSTER
        assert qn_dict['table_name'] == TB

    def test_make_table_qn(self):
        qn = make_table_qualified_name(TB, CL, DB)
        assert qn == '{}.{}@{}'.format(DB, TB, CL)

    def test_make_table_qn_with_default_database(self):
        qn = make_table_qualified_name(TB, CL, DEFAULT_DB_CLUSTER)
        assert qn == '{}@{}'.format(TB, CL)

    def test_make_table_qn_with_default_cluster(self):
        qn = make_table_qualified_name(TB, DEFAULT_DB_CLUSTER, DB)
        assert qn == '{}.{}'.format(DB, TB)

    def test_make_table_qn_only_table_with_default_db_cluster(self):
        qn = make_table_qualified_name(TB, DEFAULT_DB_CLUSTER, DEFAULT_DB_CLUSTER)
        assert qn == '{}'.format(TB)

    def test_make_table_qn_without_database(self):
        qn = make_table_qualified_name(TB, CL)
        assert qn == '{}@{}'.format(TB, CL)

    def test_make_table_qn_without_cluster(self):
        qn = make_table_qualified_name(TB, db=DB)
        assert qn == '{}.{}'.format(DB, TB)

    def test_make_table_qn_only_table(self):
        qn = make_table_qualified_name(TB)
        assert qn == '{}'.format(TB)



from atlasclient.client import Atlas


class TestClient():

    def test_atlas_client(self):
        client = Atlas('localhost', port=21000, username='admin', password='admin')
        assert client.base_url  == 'http://localhost:21000'
        assert 'headers' in client.client.request_params.keys()
        assert 'X-Requested-By' in client.client.request_params['headers']
        assert 'entity_post' in dir(client) 

    def test_http_client(self):
        pass



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from collections import namedtuple
import logging
import inspect

from atlasclient.utils import NullHandler

LOG = logging.getLogger(__name__)
LOG.addHandler(NullHandler())

EVENT_HANDLERS = {}
state_list = ['ANY', 'STARTED', 'FAILED', 'FINISHED', 'PROGRESS']
states = namedtuple('EventStates', state_list)(*state_list)


def evented(method):
    def replacement(self, *args, **kwargs):
        publish(self, method.__name__, states.STARTED)
        try:
            return method(self, *args, **kwargs)
        except Exception:
            publish(self, method.__name__, states.FAILED)
            raise
        finally:
            publish(self, method.__name__, states.FINISHED)

    return replacement


def publish(obj, event, event_state, **kwargs):
    """Publish an event from an object.

    This is a really basic pub-sub event system to allow for tracking progress
    on methods externally.  It fires the events for the first match it finds in
    the object hierarchy, going most specific to least.  If no match is found
    for the exact event+event_state, the most specific event+ANY is fired
    instead.

    Multiple callbacks can be bound to the event+event_state if desired.  All
    will be fired in the order they were registered.
    """
    # short-circuit if nothing is listening
    if len(EVENT_HANDLERS) == 0:
        return

    if inspect.isclass(obj):
        pub_cls = obj
    else:
        pub_cls = obj.__class__
    potential = [x.__name__ for x in inspect.getmro(pub_cls)]

    # if we don't find a match for this event/event_state we fire the events
    # for this event/ANY instead for the closest match
    fallbacks = None
    callbacks = []
    for cls in potential:
        event_key = '.'.join([cls, event, event_state])
        backup_key = '.'.join([cls, event, states.ANY])
        if event_key in EVENT_HANDLERS:
            callbacks = EVENT_HANDLERS[event_key]
            break
        elif fallbacks is None and backup_key in EVENT_HANDLERS:
            fallbacks = EVENT_HANDLERS[backup_key]

    if fallbacks is not None:
        callbacks = fallbacks

    for callback in callbacks:
        callback(obj, **kwargs)
    return


def subscribe(obj, event, callback, event_state=None):
    """Subscribe an event from an class.

    Subclasses of the class/object will also fire events for this class,
    unless a more specific event exists.
    """
    if inspect.isclass(obj):
        cls = obj.__name__
    else:
        cls = obj.__class__.__name__

    if event_state is None:
        event_state = states.ANY

    event_key = '.'.join([cls, event, event_state])
    if event_key not in EVENT_HANDLERS:
        EVENT_HANDLERS[event_key] = []

    EVENT_HANDLERS[event_key].append(callback)
    return



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""

from setuptools import setup, find_packages
import os
import sys

here = os.path.abspath(os.path.dirname(__file__))
sys.path.insert(0, here)
from atlasclient import __version__ as version


with open(os.path.join(here, 'README.rst')) as readme_file:
    readme = readme_file.read()

with open(os.path.join(here, 'HISTORY.rst')) as history_file:
    history = history_file.read()

requirements_path = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'requirements.txt')
test_requirements_path = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'test_requirements.txt')
setup_requirements_path = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'setup_requirements.txt')

with open(requirements_path) as requirements_file:
    requirements = requirements_file.readlines()

with open(test_requirements_path) as test_requirements_file:
    test_requirements = test_requirements_file.readlines()

with open(setup_requirements_path) as setup_requirements_file:
    setup_requirements = setup_requirements_file.readlines()

setup_args = dict(
    name='pyatlasclient',
    version=version,
    description="Apache Atlas Python Client",
    long_description=readme + '\n\n' + history,
    author="Verdan Mahmood",
    author_email='verdan.mahmood@gmail.com',
    url='https://github.com/verdan/pyatlasclient',
    packages=find_packages(include=['atlasclient']),
    include_package_data=True,
    install_requires=requirements,
    license='Apache Software License 2.0',
    zip_safe=False,
    keywords='atlasclient, pyatlasclient, apache atlas, atlas',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Natural Language :: English',
        "Programming Language :: Python :: 2",
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
    test_suite='tests',
    tests_require=test_requirements,
)

setup(**setup_args)



from pytest_mock import mocker
from pkg_resources import resource_filename
import json
import pytest

from atlasclient import client
from atlasclient import exceptions
GUID = '8bbea92b-d98c-4613-ae6e-1a9d0b4f344b'
RESPONSE_JSON_DIR = 'response_json'
QUERY_JSON_DIR = 'query_json'

@pytest.fixture(scope='class')
def entity_post_response():
    with open('{}/entity_post.json'.format(resource_filename(__name__, RESPONSE_JSON_DIR))) as json_data:
            response = json.load(json_data)
    return(response)

@pytest.fixture(scope='class')
def entity_bulk_response():
    with open('{}/entitybulk_get.json'.format(resource_filename(__name__, RESPONSE_JSON_DIR))) as json_data:
            response = json.load(json_data)
    return(response)

@pytest.fixture(scope='class')
def entity_bulk_classification_response():
    with open('{}/entitybulkclassification_post.json'.format(resource_filename(__name__, RESPONSE_JSON_DIR))) as json_data:
            response = json.load(json_data)
    return(response)

@pytest.fixture(scope='class')
def entity_guid_response():
    with open('{}/entityguid_get.json'.format(resource_filename(__name__, RESPONSE_JSON_DIR))) as json_data:
            response = json.load(json_data)
    return(response)

@pytest.fixture(scope='class')
def entity_guid_delete_response():
    with open('{}/entityguid_delete.json'.format(resource_filename(__name__, RESPONSE_JSON_DIR))) as json_data:
            response = json.load(json_data)
    return(response)

@pytest.fixture(scope='function')
def entity_guid(atlas_client):
    entity_guid = atlas_client.entity_guid(GUID) 
    return(entity_guid)

@pytest.fixture(scope='class')
def entity_guid_classification_response():
    with open('{}/entityguid_classification_get.json'.format(resource_filename(__name__, RESPONSE_JSON_DIR))) as json_data:
            response = json.load(json_data)
    return(response)

@pytest.fixture(scope='class')
def entity_guid_classifications_post():
    with open('{}/entityguid_classifications_post.json'.format(resource_filename(__name__, QUERY_JSON_DIR))) as json_data:
            response = json.load(json_data)
    return(response)

@pytest.fixture(scope='class')
def typedefs_response():
    with open('{}/typedefs_get.json'.format(resource_filename(__name__, RESPONSE_JSON_DIR))) as json_data:
            response = json.load(json_data)
    return(response)

@pytest.fixture(scope='class')
def typedefs_headers_response():
    with open('{}/typedefs_headers_get.json'.format(resource_filename(__name__, RESPONSE_JSON_DIR))) as json_data:
            response = json.load(json_data)
    return(response)

@pytest.fixture(scope='class')
def relationshipdef_guid_response():
    with open('{}/relationshipdef_guid_get.json'.format(resource_filename(__name__, RESPONSE_JSON_DIR))) as json_data:
            response = json.load(json_data)
    return(response)

@pytest.fixture(scope='class')
def lineage_guid_response():
    with open('{}/lineage_guid_get.json'.format(resource_filename(__name__, RESPONSE_JSON_DIR))) as json_data:
            response = json.load(json_data)
    return(response)

@pytest.fixture(scope='class')
def relationship_guid_response():
    with open('{}/relationship_guid_get.json'.format(resource_filename(__name__, RESPONSE_JSON_DIR))) as json_data:
            response = json.load(json_data)
    return(response)

@pytest.fixture(scope='class')
def search_attribute_response():
    with open('{}/search_attribute_get.json'.format(resource_filename(__name__, RESPONSE_JSON_DIR))) as json_data:
            response = json.load(json_data)
    return(response)

@pytest.fixture(scope='class')
def search_saved_response():
    with open('{}/search_saved_get.json'.format(resource_filename(__name__, RESPONSE_JSON_DIR))) as json_data:
            response = json.load(json_data)
    return(response)

@pytest.fixture(scope='class')
def search_saved_name_response():
    with open('{}/search_saved_name_get.json'.format(resource_filename(__name__, RESPONSE_JSON_DIR))) as json_data:
            response = json.load(json_data)
    return(response)


@pytest.fixture(scope='class')
def search_saved_guid_response():
    with open('{}/search_saved_guid_get.json'.format(resource_filename(__name__, RESPONSE_JSON_DIR))) as json_data:
            response = json.load(json_data)
    return(response)


@pytest.fixture(scope='class')
def search_saved_update():
    with open('{}/search_saved_update.json'.format(resource_filename(__name__, RESPONSE_JSON_DIR))) as json_data:
            response = json.load(json_data)
    return(response)


@pytest.fixture(scope='class')
def admin_metrics_response():
    with open('{}/admin_metrics.json'.format(resource_filename(__name__, RESPONSE_JSON_DIR))) as json_data:
            response = json.load(json_data)
    return response


class TestEntityREST():
    def test_entity_post(self, mocker, atlas_client, entity_guid_response, entity_post_response):
        mocker.patch.object(atlas_client.client, 'get')
        atlas_client.client.get.return_value = entity_guid_response
        mocker.patch.object(atlas_client.entity_post.client, 'post')
        atlas_client.entity_post.client.post.return_value = entity_post_response
        atlas_client.entity_post.create(data=entity_guid_response)
        atlas_client.entity_post.client.post.assert_called_with(atlas_client.entity_post.url, data=entity_guid_response)

    def test_entity_bulk_get(self, mocker, atlas_client, entity_bulk_response):
        mocker.patch.object(atlas_client.client, 'get')
        atlas_client.entity_bulk.client.get.return_value =  entity_bulk_response 
        params = {'guid': [GUID, '92b3a92b-d98c-4613-ae6e-1a9d0b4f344b']}
        bulk_collection = atlas_client.entity_bulk(**params) 
        for bulk in bulk_collection:
            atlas_client.entity_bulk.client.get.assert_called_with(bulk_collection.url, params=params)
            for entity in bulk.entities:
                assert entity.version == 12345

    def test_entity_bulk_get_with_relationships(self, mocker, atlas_client, entity_bulk_response):
        mocker.patch.object(atlas_client.client, 'get')
        atlas_client.entity_bulk.client.get.return_value =  entity_bulk_response
        params = {'guid': [GUID, '92b3a92b-d98c-4613-ae6e-1a9d0b4f344b']}
        bulk_collection = atlas_client.entity_bulk(**params)
        for bulk in bulk_collection:
            atlas_client.entity_bulk.client.get.assert_called_with(bulk_collection.url, params=params)
            for entity in bulk.entities_with_relationships():
                assert entity.version == 12345

    def test_entity_bulk_delete(self, mocker, atlas_client, entity_bulk_response):
        mocker.patch.object(atlas_client.client, 'get')
        atlas_client.entity_bulk.client.get.return_value =  entity_bulk_response 
        mocker.patch.object(atlas_client.client, 'delete')
        atlas_client.entity_bulk.client.delete.return_value =  entity_bulk_response 
        params = {'guid': [GUID, '92b3a92b-d98c-4613-ae6e-1a9d0b4f344b']}
        bulk_collection = atlas_client.entity_bulk(**params) 
        for bulk in bulk_collection:
            bulk.delete(**params)
            atlas_client.entity_bulk.client.delete.assert_called_with(bulk_collection.url, params=params)
    
    def test_entity_bulk_create(self, mocker, atlas_client, entity_bulk_response):
        mocker.patch.object(atlas_client.client, 'get')
        atlas_client.entity_bulk.client.get.return_value =  entity_bulk_response 
        mocker.patch.object(atlas_client.client, 'post')
        atlas_client.entity_bulk.client.post.return_value =  entity_bulk_response 
        params = {'guid': [GUID, '92b3a92b-d98c-4613-ae6e-1a9d0b4f344b']}
        bulk_collection = atlas_client.entity_bulk(**params) 
        for bulk in bulk_collection:
            bulk.create()
            atlas_client.entity_bulk.client.post.assert_called_with(bulk_collection.url, data=bulk._data)
    
    def test_entity_bulk_classification_create(self, mocker, atlas_client, entity_bulk_classification_response):
        mocker.patch.object(atlas_client.entity_bulk_classification.client, 'post')
        atlas_client.entity_bulk_classification.client.post.return_value =  entity_bulk_classification_response 
        params = {'guid': [GUID, '92b3a92b-d98c-4613-ae6e-1a9d0b4f344b']}
        atlas_client.entity_bulk_classification.create(data=entity_bulk_classification_response)
        atlas_client.entity_bulk_classification.client.post.assert_called_with(atlas_client.entity_bulk_classification.url, data=entity_bulk_classification_response) 
        
    def test_get_entity_by_guid(self, mocker, entity_guid_response, entity_guid):
        mocker.patch.object(entity_guid.client.client, 'request')
        entity_guid.client.client.request.return_value = entity_guid_response
        assert '/entity/guid/{}'.format(GUID) in entity_guid._href
        #  Loading the data because before that no data were actually loaded (lazy loading) 
        entity_guid.entity
        assert entity_guid._data['entity']['guid'] == GUID
        entity_guid.entity['status'] = 'ACTIVE'
        assert entity_guid._data['entity']['status'] == 'ACTIVE'
        
    def test_update_entity_by_guid(self, mocker, entity_guid_response, entity_guid):    
        mocker.patch.object(entity_guid.client, 'request')
        entity_guid.client.request.return_value = entity_guid_response
        mocker.patch.object(entity_guid.client, 'put')
        attribute = 'description'
        entity_guid.update(attribute=attribute)
        entity_guid.client.put.assert_called_with(entity_guid._href + '?name={}'.format(attribute),
                                                  data=entity_guid.entity['attributes'][attribute])
    
    def test_create_entity_by_guid(self, mocker, entity_guid_response, entity_guid):    
        mocker.patch.object(entity_guid.client, 'post')
        entity_guid.create()
        entity_guid.client.post.assert_called_with(entity_guid._href, data=entity_guid._data)
        
    def test_delete_entity_by_guid(self, mocker, entity_guid_delete_response, entity_guid):
        mocker.patch.object(entity_guid.client.client, 'request')
        entity_guid.client.client.request.return_value = entity_guid_response
        mocker.patch.object(entity_guid.client, 'delete')
        entity_guid.client.delete.return_value = entity_guid_delete_response
        entity_guid.delete()
        entity_guid.client.delete.assert_called_with(entity_guid._href)
        
    def test_get_entity_by_guid_classifications(self, mocker, entity_guid_classification_response, entity_guid): 
        mocker.patch.object(entity_guid.classifications.client, 'get')
        entity_guid.classifications.client.get.return_value = entity_guid_classification_response
        for c in entity_guid.classifications:
            assert 'entity/guid/{guid}/classifications'.format(guid=GUID)
    
    def test_create_entity_by_guid_classifications(self, mocker, entity_guid_classifications_post, entity_guid):
        mocker.patch.object(entity_guid.classifications.client, 'post')
        entity_guid.classifications.client.post.return_value = {}
        entity_guid.classifications.create(data=entity_guid_classifications_post)
        entity_guid.classifications.client.post.assert_called_with(entity_guid.classifications.url, data=entity_guid_classifications_post)
    
    def test_update_entity_by_guid_classifications(self, mocker, entity_guid_classification_response, entity_guid_classifications_post, entity_guid): 
        mocker.patch.object(entity_guid.classifications.client, 'get')
        entity_guid.classifications.client.get.return_value = entity_guid_classification_response
        mocker.patch.object(entity_guid.classifications.client, 'put')
        entity_guid.classifications.client.put.return_value = {}
        entity_guid.classifications.update()
        entity_guid.classifications.client.put.assert_called_with(entity_guid.classifications.url, 
                                                                  data=[{'typeName': 'Confidential'}, {'typeName': None}])
        
         
class TestTypeDefs():
    def test_typedefs_get(self, mocker, atlas_client, typedefs_response):
        mocker.patch.object(atlas_client.client, 'request')
        atlas_client.client.request.return_value = typedefs_response
        #  structDefs
        attribute_defs_names = []
        constraint_counter = 0
        for t in atlas_client.typedefs:
            for type_struct in t.structDefs:
                assert hasattr(type_struct, 'name')
                for attribute_def in type_struct.attributeDefs:
                    attribute_defs_names.append(attribute_def.name)
                    for constraint in attribute_def.constraints:
                        constraint_counter += 1
        assert attribute_defs_names == ['attributeName{}'.format(i) for i in range(1,5)]
        assert constraint_counter == 8
        #  enumDefs
        element_defs_values = []
        for t in atlas_client.typedefs:
            for type_class in t.enumDefs:
                assert hasattr(type_class, 'name')
                for element_def in type_class.elementDefs:
                    element_defs_values.append(element_def.value)
        assert element_defs_values == ['...', '...', '...', '...']
        #  classificationDefs
        class_attr_defs_values = []
        supertype_counter = 0
        for t in atlas_client.typedefs:
            for type_class in t.classificationDefs:
                assert hasattr(type_class, 'name')
                for attribute_def in type_class.attributeDefs:
                    class_attr_defs_values.append(attribute_def.name)
                for supertype in type_class.superTypes:
                    supertype_counter += 1

        assert class_attr_defs_values == ['...', '...', '...', '...']
        assert supertype_counter == 4 
        #  entityDefs
        entity_attr_defs_values = []
        supertype_counter = 0
        for t in atlas_client.typedefs:
            for type_entity in t.entityDefs:
                assert hasattr(type_entity, 'name')
                for attribute_def in type_entity.attributeDefs:
                    entity_attr_defs_values.append(attribute_def.name)
                for supertype in type_entity.superTypes:
                    supertype_counter += 1

        assert entity_attr_defs_values == ['...', '...', '...', '...']
        assert supertype_counter == 4 

    def test_typedefs_put(self, mocker, atlas_client, typedefs_response):
        mocker.patch.object(atlas_client.client, 'get')
        atlas_client.client.get.return_value = typedefs_response
        for t in atlas_client.typedefs:
                mocker.patch.object(t.client, 'put')
                t.client.put.return_value = typedefs_response 
                t.update()
                t.client.put.assert_called_with(t._href, data=t._data)

    def test_typedefs_post(self, mocker, atlas_client, typedefs_response):
        mocker.patch.object(atlas_client.client, 'get')
        atlas_client.client.get.return_value = typedefs_response
        for t in atlas_client.typedefs:
                mocker.patch.object(t.client, 'post')
                t.client.post.return_value = typedefs_response 
                t.create()
                t.client.post.assert_called_with(t._href, data=t._data)
    
    def test_typedefs_delete(self, mocker, atlas_client, typedefs_response):
        mocker.patch.object(atlas_client.client, 'get')
        atlas_client.client.get.return_value = typedefs_response
        for t in atlas_client.typedefs:
                mocker.patch.object(t.client, 'delete')
                t.client.delete.return_value = {} 
                t.delete()
                t.client.delete.assert_called_with(t._href, data=typedefs_response)

    def test_typedefs_headers_get(self, mocker, atlas_client, typedefs_headers_response):
        mocker.patch.object(atlas_client.client, 'request')
        atlas_client.client.request.return_value = typedefs_headers_response
        for h in atlas_client.typedefs_headers:
            assert h.category == 'RELATIONSHIP'
           
    def test_classificationdef_guid_get(self, mocker, atlas_client, typedefs_response):
        mocker.patch.object(atlas_client.client, 'request')
        atlas_client.client.request.return_value = typedefs_response['classificationDefs'][0]
        c_def = atlas_client.classificationdef_guid(GUID)
        assert c_def.category == 'CLASSIFICATION'
   
    def test_classificationdef_name_get(self, mocker, atlas_client, typedefs_response):
        mocker.patch.object(atlas_client.client, 'request')
        atlas_client.client.request.return_value = typedefs_response['classificationDefs'][0]
        c_def = atlas_client.classificationdef_name('name')
        assert c_def.category == 'CLASSIFICATION'

    def test_entitydef_guid_get(self, mocker, atlas_client, typedefs_response):
        mocker.patch.object(atlas_client.client, 'request')
        atlas_client.client.request.return_value = typedefs_response['entityDefs'][0]
        c_def = atlas_client.entitydef_guid(GUID)
        assert c_def.category == 'ENTITY'
   
    def test_entitydef_name_get(self, mocker, atlas_client, typedefs_response):
        mocker.patch.object(atlas_client.client, 'request')
        atlas_client.client.request.return_value = typedefs_response['entityDefs'][0]
        c_def = atlas_client.entitydef_name('name')
        assert c_def.category == 'ENTITY'

    def test_enumdef_guid_get(self, mocker, atlas_client, typedefs_response):
        mocker.patch.object(atlas_client.client, 'request')
        atlas_client.client.request.return_value = typedefs_response['enumDefs'][0]
        c_def = atlas_client.enumdef_guid(GUID)
        assert c_def.category == 'ENUM'
   
    def test_enumdef_name_get(self, mocker, atlas_client, typedefs_response):
        mocker.patch.object(atlas_client.client, 'request')
        atlas_client.client.request.return_value = typedefs_response['enumDefs'][0]
        c_def = atlas_client.enumdef_name('name')
        assert c_def.category == 'ENUM'
    
    def test_relationshipdef_guid_get(self, mocker, atlas_client, relationshipdef_guid_response):
        mocker.patch.object(atlas_client.client, 'request')
        atlas_client.client.request.return_value = relationshipdef_guid_response
        c_def = atlas_client.relationshipdef_guid(GUID)
        assert c_def.relationshipCategory == 'ASSOCIATION'
   
    def test_relationshipdef_name_get(self, mocker, atlas_client, relationshipdef_guid_response):
        mocker.patch.object(atlas_client.client, 'request')
        atlas_client.client.request.return_value = relationshipdef_guid_response
        c_def = atlas_client.relationshipdef_name('name')
        assert c_def.relationshipCategory == 'ASSOCIATION'
    
    def test_structdef_guid_get(self, mocker, atlas_client, typedefs_response):
        mocker.patch.object(atlas_client.client, 'request')
        atlas_client.client.request.return_value = typedefs_response['structDefs'][0]
        c_def = atlas_client.structdef_guid(GUID)
        assert c_def.category == 'STRUCT'
   
    def test_structdef_name_get(self, mocker, atlas_client, typedefs_response):
        mocker.patch.object(atlas_client.client, 'request')
        atlas_client.client.request.return_value = typedefs_response['structDefs'][0]
        c_def = atlas_client.structdef_name('name')
        assert c_def.category == 'STRUCT'

    def test_typedef_guid_get(self, mocker, atlas_client, typedefs_response):
        mocker.patch.object(atlas_client.client, 'request')
        atlas_client.client.request.return_value = typedefs_response['structDefs'][0]
        c_def = atlas_client.typedef_guid(GUID)
        assert c_def.category == 'STRUCT'
   
    def test_typedef_name_get(self, mocker, atlas_client, typedefs_response):
        mocker.patch.object(atlas_client.client, 'request')
        atlas_client.client.request.return_value = typedefs_response['structDefs'][0]
        c_def = atlas_client.typedef_name('name')
        assert c_def.category == 'STRUCT'

class TestLineageGuid():
    def test_lineage_guid_get(self, mocker, atlas_client, lineage_guid_response):
        mocker.patch.object(atlas_client.client, 'request')
        atlas_client.client.request.return_value =  lineage_guid_response 
        lineage = atlas_client.lineage_guid(GUID)
        assert lineage.lineageDirection == 'BOTH'

class TestRelationshipREST():
    def test_relationship_guid_get(self, mocker, atlas_client, relationship_guid_response):
        mocker.patch.object(atlas_client.client, 'request')
        atlas_client.client.request.return_value =  relationship_guid_response 
        relationship = atlas_client.relationship_guid(GUID)
        assert relationship.status == 'ACTIVE'

    def test_relationship_guid_delete(self, mocker, atlas_client, relationship_guid_response):
        mocker.patch.object(atlas_client.client, 'request')
        atlas_client.client.request.return_value =  relationship_guid_response 
        relationship = atlas_client.relationship_guid(GUID)
        relationship.delete()
 
    def test_relationship_guid_update(self, mocker, atlas_client, relationship_guid_response):
        mocker.patch.object(atlas_client.client, 'request')
        atlas_client.client.request.return_value =  relationship_guid_response 
        relationship = atlas_client.relationship_guid(GUID)
        relationship.update()
    
    def test_relationship_guid_create(self, mocker, atlas_client, relationship_guid_response):
        mocker.patch.object(atlas_client.client, 'request')
        atlas_client.client.request.return_value =  relationship_guid_response 
        relationship = atlas_client.relationship_guid(GUID)
        with pytest.raises(exceptions.MethodNotImplemented) as e:
            relationship.create()

#    def test_relationship_get(self, mocker, atlas_client, relationship_guid_response):
#        with pytest.raises(exceptions.MethodNotImplemented) as e: 
#            atlas_client.relationship.refresh() 

    def test_relationship_put(self, mocker, atlas_client, relationship_guid_response):
        relationship_collection = atlas_client.relationship(data=relationship_guid_response) 
        for r in relationship_collection:
            assert r._data == relationship_guid_response
            mocker.patch.object(r.client, 'put')
            r.client.put.return_value = relationship_guid_response 
            r.update()
            r.client.put.assert_called_with(r._href, data=r._data)
             
    def test_relationship_post(self, mocker, atlas_client, relationship_guid_response):
        relationship_collection = atlas_client.relationship(data=relationship_guid_response) 
        for r in relationship_collection:
            assert r._data == relationship_guid_response
            mocker.patch.object(r.client, 'post')
            r.client.post.return_value = relationship_guid_response 
            r.create()
            r.client.post.assert_called_with(r._href, data=r._data)
        

class TestDiscoveryREST():
    def test_search_attribute_get(self, mocker, atlas_client, search_attribute_response):
        mocker.patch.object(atlas_client.search_attribute.client, 'get')
        atlas_client.search_attribute.client.get.return_value =  search_attribute_response 
        params = {'attrName': 'attrName', 'attrValue': 'attrVal', 'offset': '1'}
        search_results = atlas_client.search_attribute(**params) 
        for s in search_results:
            assert s.queryType == 'GREMLIN'
            atlas_client.search_attribute.client.get.assert_called_with(s.url, params=params)
        for s in search_results:
            for e in s.entities:
                assert e.attributes['property1'] == {}

    def test_search_basic_get(self, mocker, atlas_client, search_attribute_response):
        mocker.patch.object(atlas_client.search_basic.client, 'get')
        search_attribute_response['queryType'] = 'BASIC'
        atlas_client.search_basic.client.get.return_value = search_attribute_response
        params = {'classification': 'class', 'excludedDeletedEntities': 'True', 'limit': '1'}
        search_results = atlas_client.search_basic(**params) 
        for s in search_results:
            assert s.queryType == 'BASIC'
            atlas_client.search_basic.client.get.assert_called_with(s.url, params=params)
            for e in s.entities:
                assert e.attributes['property1'] == {}

    def test_search_basic_post(self, mocker, atlas_client, search_attribute_response):
        mocker.patch.object(atlas_client.search_basic.client, 'get')
        search_attribute_response['queryType'] = 'BASIC'
        atlas_client.search_basic.client.get.return_value =  search_attribute_response 
        mocker.patch.object(atlas_client.search_basic.client, 'post')
        atlas_client.search_basic.client.post.return_value =  search_attribute_response
        search_results = atlas_client.search_basic
        for s in search_results:
            assert s.queryType == 'BASIC'
            s.create(data=s._data)
            atlas_client.search_basic.client.post.assert_called_with(s.url, data=s._data)

    def test_search_dsl_get(self, mocker, atlas_client, search_attribute_response):
        mocker.patch.object(atlas_client.search_dsl.client, 'get')
        search_attribute_response['queryType'] = 'DSL'
        atlas_client.search_dsl.client.get.return_value =  search_attribute_response 
        params = {'classification': 'class', 'limit': '1'}
        search_results = atlas_client.search_dsl(**params) 
        for s in search_results:
            assert s.queryType == 'DSL'
            atlas_client.search_dsl.client.get.assert_called_with(s.url, params=params)
            for e in s.entities:
                assert e.attributes['property1'] == {}
            assert s.flatten_attrs() == ['12', '34', '56']

    def test_search_fulltext_get(self, mocker, atlas_client, search_attribute_response):
        mocker.patch.object(atlas_client.search_fulltext.client, 'get')
        search_attribute_response['queryType'] = 'ATTRIBUTE'
        atlas_client.search_fulltext.client.get.return_value =  search_attribute_response 
        params = {'classification': 'class', 'excludedDeletedEntities': 'True', 'limit': '1'}
        search_results = atlas_client.search_fulltext(**params) 
        for s in search_results:
            assert s.queryType == 'ATTRIBUTE'
            atlas_client.search_fulltext.client.get.assert_called_with(s.url, params=params)
            for e in s.entities:
                assert e.attributes['property1'] == {}

    def test_search_saved_get(self, mocker, atlas_client, search_saved_response):
        mocker.patch.object(atlas_client.search_saved.client, 'get')
        atlas_client.search_saved.client.get.return_value = search_saved_response
        search_results = atlas_client.search_saved()
        for s in search_results:
            assert s.name == 'TESTNAME'

    def test_search_saved_name_get(self, mocker, atlas_client, search_saved_name_response):
        mocker.patch.object(atlas_client.search_saved.client, 'request')
        atlas_client.search_saved.client.request.return_value = search_saved_name_response
        result = atlas_client.search_saved('TESTNAME')

        assert result.name == 'TESTNAME'

    def test_search_saved_guid_delete(self, mocker, atlas_client, search_saved_guid_response):
        mocker.patch.object(atlas_client.client, 'request')
        atlas_client.client.request.return_value = search_saved_guid_response
        response = atlas_client.search_saved('TESTNAME')
        response.delete(guid=GUID)

    def test_search_saved_create(self, mocker, atlas_client, search_saved_name_response):
        mocker.patch.object(atlas_client.search_saved.client, 'post')
        atlas_client.search_saved.client.post.return_value = search_saved_response
        atlas_client.search_saved.create(data=search_saved_response)

    def test_search_saved_update(self, mocker, atlas_client, search_saved_guid_response, search_saved_update):
        mocker.patch.object(atlas_client.search_saved.client, 'request')
        atlas_client.search_saved.client.request.return_value = search_saved_name_response
        result = atlas_client.search_saved('TESTNAME')

        mocker.patch.object(atlas_client.search_saved.client, 'put')
        atlas_client.search_saved.client.put.return_value = search_saved_update

        result.update(data=search_saved_update)


class TestAdminREST:
    def test_admin_metrics_get(self, mocker, atlas_client, admin_metrics_response):
        mocker.patch.object(atlas_client.client, 'request')
        atlas_client.client.request.return_value = admin_metrics_response

        for metrics in atlas_client.admin_metrics:
            assert hasattr(metrics, 'entity')
            assert hasattr(metrics, 'general')
            assert hasattr(metrics, 'tag')

            entity_stats = metrics.entity
            assert entity_stats['entityDeleted'] == {}
            assert len(entity_stats['entityActive'].keys()) == 9

            general_stats = metrics.general
            assert general_stats['collectionTime'] == 1565123782025
            assert general_stats['entityCount'] == 46

            tag_stats = metrics.tag
            assert tag_stats['tagEntities']['JdbcAccess'] == 2
            assert len(tag_stats['tagEntities'].keys()) == 7
